// Copyright 1998, Ben Goetter.  All rights reserved.

/* stdio.cpp */

#if (defined(WIN32_PLATFORM_PSPC) && (_WIN32_WCE >= 300)) || defined(x86)
#else

#include "stdinc.h"
#include <limits.h>
#define SZ_DEFAULT_CAPTION __T("STDIO")
//#include "assert.h"

#define IMPLEMENTS_STDIO
#include "stdio.h"
#include "stdioimp.h"


#define MAXOPENFILES 10
#define assert(x)	1
// These two have streams defined and owned by the client
FILE StdioStdin = {FALSE, NULL};
FILE StdioStdout = {FALSE, NULL};
// These all have streams owned by stdio itself
FILE RgStdioFiles[MAXOPENFILES] = {
	{FALSE, NULL},
	{FALSE, NULL},
	{FALSE, NULL},
	{FALSE, NULL},
	{FALSE, NULL},
	{FALSE, NULL},
	{FALSE, NULL},
	{FALSE, NULL},
	{FALSE, NULL},
	{FALSE, NULL}
};


inline BOOL IsInvalidFile(FILE* pfile)
{
	return ((NULL == pfile) || (!pfile->fActive));
}


// DLL entry point

HANDLE g_hinstDll = NULL;

BOOL WINAPI DllMain(HANDLE hinstDll, DWORD fdwReason, LPVOID)
{
 	if (DLL_PROCESS_ATTACH == fdwReason)
 	{
		g_hinstDll = hinstDll;
 	}
	else if (DLL_PROCESS_DETACH == fdwReason)
	{
		for (int i = 0; i < MAXOPENFILES; ++i)
			if (RgStdioFiles[i].fActive)
				delete RgStdioFiles[i].pstream;
	}

 	return TRUE;
}


//$ BUGBUG - this package needs errno/strerror, 
// or some other way of getting an errorcode out of the stdio-ish interface
// Also - support _fmode, ferror, _doserrno


// A small buffer 
// that allocates itself from the frame whenever possible

class CLocalTempBuffer
{
private:
	BYTE _rgbFrame[256];
	BYTE* _pbAlloc;

public:
	CLocalTempBuffer() : _pbAlloc(NULL) {}
	~CLocalTempBuffer()
		{ if (_pbAlloc) ::LocalFree(_pbAlloc); }

	BOOL Init(unsigned cb)
	{
		assert(NULL == _pbAlloc);
		if (cb > sizeof(_rgbFrame))
		{
			_pbAlloc = (BYTE*)::LocalAlloc(LMEM_FIXED, cb);
			if (NULL == _pbAlloc)
				return FALSE;
		}
		return TRUE;
	}

	BYTE* GetBuffer() const
		{ return ((_pbAlloc == NULL) ? (BYTE*)_rgbFrame : _pbAlloc); }
};



// Exported to main wndproc holder - called on WM_HIBERNATE

void StdioOnHibernate()
{
	//$ BUGBUG - may be called from UI thread while streams in use by SIOD
	// should guard with critical section

	if (StdioStdin.fActive)
		StdioStdin.pstream->OnHibernate();
	if (StdioStdout.fActive)
		StdioStdout.pstream->OnHibernate();
	for (int i = 0; i < MAXOPENFILES; ++i)
		if (RgStdioFiles[i].fActive)
			RgStdioFiles[i].pstream->OnHibernate();
}


void StdioInitStdin(CXStdioStream* pstreamStdin)
{
	StdioStdin.pstream = pstreamStdin;
	StdioStdin.fActive = (NULL != pstreamStdin);
}


void StdioInitStdout(CXStdioStream* pstreamStdout)
{
	StdioStdout.pstream = pstreamStdout;
	StdioStdout.fActive = (NULL != pstreamStdout);
}


CXStdioStream::CXStdioStream() : 
		_fPushback(FALSE), _chPushback(0), 
		_fReading(FALSE), _fWriting(FALSE), _fAppending(FALSE),
		_fCooked(FALSE)
{}


CXStdioStream::~CXStdioStream()
{}


void CXStdioStream::Flush()
{
	// Default implementation only clears pushback buffer
	// (good enough for stdin/out)
	_fPushback = FALSE;
}

unsigned CXStdioStream::Tell()
{
	// not implemented in stdin/stdout
	return 0;
}

BOOL CXStdioStream::Seek(int ibOffset, int nType)
{
	// not implemented in stdin/stdout
	return FALSE;
}

void CXStdioStream::Read(BYTE* pbIn, unsigned cbInMax, unsigned* pcbInActual)
{
	// not implemented in stdout
	*pcbInActual = 0;
}

void CXStdioStream::Write(const BYTE* pbOut, unsigned cbOut, unsigned* pcbOutActual)
{
	// not implemented in stdin
	*pcbOutActual = 0;
}

void CXStdioStream::Pushback(unsigned char ch)
{
	_chPushback = ch;
	_fPushback = TRUE;
}

int CXStdioStream::CharReady()
{
	// The default - only needs override for tty devices
	return 1;
}


CXBufferedStream::CXBufferedStream() 
	: _pbBuffer(NULL), _cbBuffer(0)
{}


CXBufferedStream::~CXBufferedStream()
{
	if (_pbBuffer)
		::LocalFree(_pbBuffer);
}


BOOL CXBufferedStream::RequireBuffer(unsigned cbRequired)
{
	if (cbRequired > _cbBuffer)
	{
		if (_pbBuffer)
		{
			::LocalFree(_pbBuffer);
			_pbBuffer = NULL;
			_cbBuffer = 0;
		}
		// Round to multiple of 256b as anti-heap-frag measure
		const unsigned cbAlloc = (((cbRequired-1)/256)+1) * 256;
		_pbBuffer = (BYTE*)::LocalAlloc(LMEM_FIXED, cbAlloc);
		if (!_pbBuffer)
		{
			return FALSE; // Signal EOF
		}
		_cbBuffer = cbAlloc;
	}
	return TRUE;
}


void CXBufferedStream::ReleaseBuffer()
{
	if (_pbBuffer)
	{
		::LocalFree(_pbBuffer);
		_pbBuffer = NULL;
		_cbBuffer = 0;
	}
}


CXBufferedReadStream::CXBufferedReadStream() : 
#if 0 // for stdin Seek/Tell - nyi
		_cbReadTotal(0), 
#endif
		_cbActual(0), _ibCurrent(0)
{}


CXBufferedReadStream::~CXBufferedReadStream()
{}


void CXBufferedReadStream::OnHibernate()
{
	if (_cbBuffer > 0 && (0 == _cbActual || _cbActual == _ibCurrent))
	{
		ResetReadBuffer();
		ReleaseBuffer();
	}
}


void CXBufferedReadStream::Read(BYTE* pbIn, unsigned cbInMax, unsigned* pcbInActual)
{
	unsigned cbReadActual = 0;

	if (_fPushback && cbInMax > 0)
	{
		*pbIn++ = _chPushback;
		_fPushback = FALSE;
		--cbInMax;
		++cbReadActual;
	}

	while (cbInMax > 0) // didn't get it all out of pushback? what a surprise
	{
		if (cbInMax > 0 && _cbActual > 0 && _ibCurrent < _cbActual)
		{
			// Get what we can out of the read-buffer

			unsigned cbAvailable = min(cbInMax, (_cbActual-_ibCurrent));
			memcpy(pbIn, _pbBuffer+_ibCurrent, cbAvailable);
			pbIn += cbAvailable;
			cbInMax -= cbAvailable;
			_ibCurrent += cbAvailable;
			cbReadActual += cbAvailable;
		}

		if (cbInMax > 0) // As necessary, reload the read-buffer
		{
			ResetReadBuffer();
			if (!LoadReadBuffer(cbInMax)) // pass remaining amt needed as hint
				break;
		}
	}

#if 0 // track for stdin/stdout Seek/Tell
	_cbReadTotal += cbReadActual;
#endif
	*pcbInActual = cbReadActual;
}


// Convert to MBCS, flattening CRLF sequences to LF as we go

BOOL CookInputTextWA(const WCHAR* pwchSrc, unsigned cchSrc, 
				     char* pchDst0, unsigned cbInAvail, unsigned* pcbEmitted,
				     BOOL* pfHadTrailingCR,
				     unsigned* pcchCRsEaten)
{
	char* pchDst = pchDst0; // retain original for pcbEmitted calc

	unsigned cchCRsEaten = 0; // count number of CRLF sequences flattened

	if (cchSrc > 0 && cbInAvail > 0)
	{
		const WCHAR* pwchEol = NULL;
		while (pwchEol = wcsstr(pwchSrc, __T("\r\n")))
		{
			const int cchLine = pwchEol - pwchSrc;
			if (cchLine > 0)
			{
				const int cbMbcs = ::WideCharToMultiByte(CP_ACP, 0, 
					pwchSrc, cchLine,
					pchDst, cbInAvail, NULL, NULL);
				if (0 == cbMbcs)
					return FALSE; // Signal EOF
				pchDst += cbMbcs;
				cbInAvail -= cbMbcs;
			}
			*pchDst++ = '\n';
			--cbInAvail;
			pwchSrc = pwchEol + 2; // CRLF
			cchSrc -= cchLine+2;
			++cchCRsEaten;
		}
	}
	if (cchSrc > 0 && cbInAvail > 0)
	{
		// Need to detect a final CR character,
		// to catch the case where a CRLF sequence crosses the boundary of a read buffer.
		// Caller can look ahead and patch the final char in the outstring as necessary

		if (NULL != pfHadTrailingCR)
			*pfHadTrailingCR = (__T('\r') == pwchSrc[cchSrc-1]);

		const int cbMbcs = ::WideCharToMultiByte(CP_ACP, 0, 
			pwchSrc, cchSrc,
			pchDst, cbInAvail, NULL, NULL);
		if (0 == cbMbcs)
			return FALSE; // Signal EOF
		pchDst += cbMbcs;
	}
	else
	{
		if (NULL != pfHadTrailingCR) 
			*pfHadTrailingCR = FALSE; // all were converted
	}

	*pcbEmitted = pchDst - pchDst0;
	if (NULL != pcchCRsEaten)
		*pcchCRsEaten = cchCRsEaten;
	return TRUE;
}


// Flatten CRLF sequences to LF - no Unicode decoding necessary

BOOL CookInputTextAA(const char* pchSrc, unsigned cchSrc, 
				     char* pchDst0, unsigned cbInAvail, unsigned* pcbEmitted,
				     BOOL* pfHadTrailingCR,
				     unsigned* pcchCRsEaten)
{
	char* pchDst = pchDst0; // retain original for pcbEmitted calc

	unsigned cchCRsEaten = 0; // count number of CRLF sequences flattened

	if (cchSrc > 0 && cbInAvail > 0)
	{
		const char* pchEol = NULL;
		while (pchEol = strstr(pchSrc, "\r\n"))
		{
			const int cchLine = pchEol - pchSrc;
			if (cchLine > 0)
			{
				const int cbCopy = min(cchLine, (int)cbInAvail);
				memcpy(pchDst, pchSrc, cbCopy);
				pchDst += cbCopy;
				cbInAvail -= cbCopy;
			}
			*pchDst++ = '\n';
			--cbInAvail;
			pchSrc = pchEol + 2; // CRLF
			cchSrc -= cchLine+2;
			++cchCRsEaten;
		}
	}
	if (cchSrc > 0 && cbInAvail > 0)
	{
		// Need to detect a final CR character,
		// to catch the case where a CRLF sequence crosses the boundary of a read buffer.
		// Caller can look ahead and patch the final char in the outstring as necessary

		if (NULL != pfHadTrailingCR)
			*pfHadTrailingCR = ('\r' == pchSrc[cchSrc-1]);

		const int cbCopy = min(cchSrc, cbInAvail);
		memcpy(pchDst, pchSrc, cbCopy);
		pchDst += cbCopy;
	}
	else
	{
		if (NULL != pfHadTrailingCR) 
			*pfHadTrailingCR = FALSE; // all were converted
	}

	*pcbEmitted = pchDst - pchDst0;
	if (NULL != pcchCRsEaten)
		*pcchCRsEaten = cchCRsEaten;
	return TRUE;
}


// Convert to Unicode, expanding newlines into CRNL sequences

BOOL CookOutputTextAW(const char* pchSrc, unsigned cbSrc, unsigned cbNewlines,
                      WCHAR* pwchDst, unsigned cchAvailable, unsigned* pcchEmitted )
{
	unsigned cchOut = 0; // count of characters written to pwchOut

	if (cbNewlines > 0 && cbSrc > 0)
	{
		// Seeking newlines in the sourcetext, one line at a time
		// Thought: since we have cbNewlines, could use strchr here,
		// decrementing a count of the # of newlines found.
		// That approach would easily transition to an approach
		// where the initial cbNewline count also saved offsets.
		const char* pch = pchSrc;
		unsigned cb = cbSrc;
		while (cb--)
		{
			if (*pch++ == '\n') //$ BUGBUG DBCS, again
			{
				// Copy/convert everything but the newline
				const int cbLine = cbSrc - cb - 1;
				if (cbLine > 0)
				{
					const int cch = ::MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, 
						pchSrc, cbLine,
						pwchDst, cchAvailable);
					if (0 == cch)
					{
						*pcchEmitted = cchOut;
						return FALSE;
					}
					cchOut += cch;
					pwchDst += cch;
					cchAvailable -= cch;
				}
				// Emit a 'cooked' newline sequence to replace the newline
				*pwchDst++ = __T('\r');
				*pwchDst++ = __T('\n');
				cchOut += 2;
				cchAvailable -= 2;

				cbSrc = cb;
				pchSrc = pch;
			}
		}
	}
	if (cbSrc > 0)
	{
		// Leftovers at the end
		const int cch = ::MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, 
			pchSrc, cbSrc,
			pwchDst, cchAvailable);
		if (0 == cch)
		{
			*pcchEmitted = cchOut;
			return FALSE;
		}
		cchOut += cch;
		pwchDst += cch;
		cchAvailable -= cch;
	}

	*pcchEmitted = cchOut;
	return TRUE;
}


// Expanding newlines into CRNL sequences, sans Unicode

BOOL CookOutputTextAA(const char* pchSrc, unsigned cbSrc, unsigned cbNewlines,
				      char* pchDst, unsigned cchAvailable, unsigned* pcchEmitted )
{
	unsigned cchOut = 0; // count of characters written to pwchOut

	if (cbNewlines > 0 && cbSrc > 0)
	{
		// Seeking newlines in the sourcetext, one line at a time
		// Thought: since we have cbNewlines, could use strchr here,
		// decrementing a count of the # of newlines found.
		// That approach would easily transition to an approach
		// where the initial cbNewline count also saved offsets.

		const char* pch = pchSrc;
		unsigned cb = cbSrc;
		while (cb--)
		{
			if (*pch++ == '\n') //$ BUGBUG DBCS, again
			{
				// Copy/convert everything but the newline
				const int cbLine = cbSrc - cb - 1;
				if (cbLine > 0)
				{
					memcpy(pchDst, pchSrc, cbLine);
					cchOut += cbLine;
					pchDst += cbLine;
					cchAvailable -= cbLine;
				}
				// Emit a 'cooked' newline sequence to replace the newline
				*pchDst++ = '\r';
				*pchDst++ = '\n';
				cchOut += 2;
				cchAvailable -= 2;

				cbSrc = cb;
				pchSrc = pch;
			}
		}
	}
	if (cbSrc > 0)
	{
		// Leftovers at the end
		memcpy(pchDst, pchSrc, cbSrc);
		cchOut += cbSrc;
		pchDst += cbSrc;
		cchAvailable -= cbSrc;
	}

	*pcchEmitted = cchOut;
	return TRUE;
}


// Count the number of newlines in source

unsigned CountNewlines(const char* pch, unsigned cb)
{
	//$ BUGBUG DBCS - sigh -
	// can't use strchr loop because pchOut not NUL terminated

	unsigned cbNewlines = 0;
	while (cb--)
		if (*pch++ == '\n')
			++cbNewlines;
	return cbNewlines;
}


CXFileStream::CXFileStream()
	: _h(NULL), _fUseUnicode(FALSE), _cchCRsEaten(0)
{
	// Since this object is reused, do all real init in Open
}


CXFileStream::~CXFileStream()
{
	Close();
}


BOOL CXFileStream::LoadReadBuffer(unsigned cbHint)
{
	const int cchReadQuantum = 512; // used instead of 'cbHint'

	if (NULL == _h)
		return FALSE;

	if (!_fCooked)
	{
		// 1 character, 1 byte. Life is simple and good.

		if (!RequireBuffer(cchReadQuantum)) // set _{c,p}bBuffer
			return FALSE;

		DWORD cbActuallyRead = 0L;
		if (!::ReadFile(_h, _pbBuffer, cchReadQuantum, &cbActuallyRead, NULL))
		{
			//$ BUGBUG better error logging?
			return FALSE;
		}
		if (0 == cbActuallyRead)
			return FALSE;
		_cbActual = cbActuallyRead;
	}
	else
	{
		// Require space for reading text from file,
		// plus space for cooking text: CRNL->NL, possibly Unicode->MBCS.

		// cbRequired: one of the (1+cchReadQuantum)*cbChar is for the WCHAR input,
		// and one is for the MBCS output,
		// pessimistically assuming that every wchar will expand into a 2-byte dbcs sequence.
		// While this seems pointless, given that dbcs input would blow SIOD out of the water,
		// that might change someday....

		const unsigned cbOneChar = _fUseUnicode ? sizeof(WCHAR) : sizeof(char);
		const unsigned cbRead = cchReadQuantum*cbOneChar;
		const unsigned cbRequired = 2*((1+cchReadQuantum)*cbOneChar);

		if (!RequireBuffer(cbRequired)) // set _{c,p}bBuffer
			return FALSE;

		// Fetch source text into tail-end of buffer

		const int ibOffset = (1+cchReadQuantum)* cbOneChar;
		BYTE* pbInPreConvert = _pbBuffer+ibOffset;
		DWORD cbActuallyRead = 0L;
		if (!::ReadFile(_h, pbInPreConvert, cbRead, &cbActuallyRead, NULL))
			return FALSE;
		if (0 == cbActuallyRead)
			return FALSE;

		BOOL fHadTrailingCR;
		unsigned cchCRsEaten;
		if (_fUseUnicode)
		{
			// Convert to MBCS, flattening CRLF sequences to LF as we go
			const unsigned cch = cbActuallyRead/sizeof(WCHAR);
			((WCHAR*)pbInPreConvert)[cch] = 0;
			if (!CookInputTextWA((const WCHAR*)pbInPreConvert, cch, 
				(char*)_pbBuffer, ibOffset, &_cbActual, &fHadTrailingCR, &cchCRsEaten))
				return FALSE;
		}
		else
		{
			// Copy and flatten
			((char*)pbInPreConvert)[cbActuallyRead] = 0;
			if (!CookInputTextAA((const char*)pbInPreConvert, cbActuallyRead, 
				(char*)_pbBuffer, ibOffset, &_cbActual, &fHadTrailingCR, &cchCRsEaten))
				return FALSE;
		}

		// Special evil case: if last character was a CR, 
		// we might have a CRLF sequence crossing the boundary of the read buffer.

		if (fHadTrailingCR)
		{
			// Peek ahead to see if it is so.
			// (Assumes we're working from a disk file. Were this a live IPC channel,
			// we'd delete the trailing CR, then save the pending-possible-CRLF
			// state to resolve in the next Read call.)

			BOOL fFoundNL;
			if (_fUseUnicode)
			{
				WCHAR chReadAhead;
				if (!::ReadFile(_h, &chReadAhead, sizeof(WCHAR), &cbActuallyRead, NULL))
					return FALSE;
				fFoundNL = (sizeof(WCHAR) == cbActuallyRead && __T('\n') == chReadAhead);
			}
			else
			{
				char chReadAhead;
				if (!::ReadFile(_h, &chReadAhead, sizeof(char), &cbActuallyRead, NULL))
					return FALSE;
				fFoundNL = (sizeof(char) == cbActuallyRead && '\n' == chReadAhead);
			}

			if (fFoundNL)
			{
				_pbBuffer[_cbActual] = '\n';
				++cchCRsEaten;
			}
			if (cbActuallyRead > 0 && !fFoundNL)
				::SetFilePointer(_h, -1*cbActuallyRead, NULL, FILE_CURRENT);
		}

		_cchCRsEaten = cchCRsEaten;
	}

	return (_cbActual > 0);
}


void CXFileStream::Write(const BYTE* pbOut, unsigned cbOut, unsigned* pcbOutActual)
{
	if (NULL == _h)
		return;

	if (_fAppending)
	{
		// All writes move to end of file
		::SetFilePointer(_h, 0, NULL, FILE_END);
	}
	else if (_fReading && _cbActual > 0)
	{
		// Write must move back to point of last "read" character
		int ibOffset = _cbActual - _ibCurrent;
		if (_fCooked && ibOffset > 0)
			ibOffset += CountNewlines((const char*)(_pbBuffer+_ibCurrent), ibOffset);
		if (_fUseUnicode && ibOffset > 0)
			ibOffset *= sizeof(WCHAR);
		if (ibOffset > 0)
			::SetFilePointer(_h, -ibOffset, NULL, FILE_CURRENT);
	}

	if (_fReading)
		ResetReadBuffer();

	if (!_fCooked)
	{
		DWORD cbActualWrite = 0L;
		if (!::WriteFile(_h, pbOut, cbOut, &cbActualWrite, NULL))
		{
			//$ BUGBUG better error logging?
			*pcbOutActual = 0;
			return;
		}
		*pcbOutActual = cbActualWrite;
	}
	else
	{
		// Use "cooked" to emit Unicode text, expand LF to CRLF.

		const unsigned cbNewlines = CountNewlines((const char*)pbOut, cbOut);
		const unsigned cbOneChar = _fUseUnicode ? sizeof(WCHAR) : sizeof(char);

		CLocalTempBuffer buf;
		if (!buf.Init( (cbOut+cbNewlines)*cbOneChar ))
		{
			*pcbOutActual = 0;
			return;
		}
		BYTE* pbOutConverted = buf.GetBuffer();

		unsigned cch = 0;
		if (_fUseUnicode)
		{
			if (!CookOutputTextAW((const char*)pbOut, cbOut, cbNewlines,
				(WCHAR*)pbOutConverted, cbOut+cbNewlines, &cch))
			{
				*pcbOutActual = 0;
				return;
			}
		}
		else
		{
			if (!CookOutputTextAA((const char*)pbOut, cbOut, cbNewlines,
				(char*)pbOutConverted, cbOut+cbNewlines, &cch))
			{
				*pcbOutActual = 0;
				return;
			}
		}

		DWORD cbActualWrite = 0L;
		if (!::WriteFile(_h, pbOutConverted, cbOneChar*cch, &cbActualWrite, NULL))
		{
			//$ BUGBUG better error logging?
			*pcbOutActual = 0;
			return;
		}
		*pcbOutActual = cbActualWrite;
	}
}


unsigned CXFileStream::TellUnbuffered()
{
	unsigned ib = ::SetFilePointer(_h, 0, NULL, FILE_CURRENT);

	if (_fUseUnicode)
	{
		ib /= sizeof(WCHAR);
		ib -= 1; // for BOM
	}

	return ib;
}


unsigned CXFileStream::Tell()
{
	if (NULL == _h)
		return 0;

	unsigned ib = TellUnbuffered();

	if (_fReading)
	{
		// Adjust for characters in read buffer
		ib -= _cbActual;
		// Adjust again for CR characters missing from read buffer
		if (_fCooked)
			ib -= _cchCRsEaten;
		// Move forward by characters "read" in read buffer
		ib += _ibCurrent;
	}

	return ib;
}


BOOL CXFileStream::Seek(int ibOffset, int nType)
{
	if (NULL == _h)
		return FALSE;

	DWORD dwMethod;
	switch (nType)
	{
	case SEEK_CUR:
		dwMethod = FILE_CURRENT;
		break;
	case SEEK_END:
		dwMethod = FILE_END;
		break;
	case SEEK_SET:
	default:
		dwMethod = FILE_BEGIN;
		break;
	}

	//
	// File offsets vs. stream buffer offsets
	// ====================================================
	//         ^         ^          ^  
	//         |         |          |
	//         |         |          |
	//         ibInFile  |          ibInFile 
	//      - _cbActual  |          (TRUE file offset)
	//   - _cchCRsEaten  |
	//                   ibInFile
	//                 - _cbActual
	//                 - _cchCRsEaten
	//                 + _ibCurrent
	//                  (VIRTUAL file offset)
	//
	// Note that for the purposes of these calculations
	// we place all the imaginary lost bytes represented by _cchCRsEaten
	// between the last buffered-read byte - _pbBuffer+_cbActual-1 - and ibInFile.
	// There is no attempt to associate a particular lost CR 
	// with a particular offset in the file.

	if (_fReading )
	{
		if (nType != SEEK_END)
		{
			// If the new offset lands within the read-buffer, "seek" takes place therein

			const unsigned ibInFile = TellUnbuffered();
			const unsigned ibStartBuf = ibInFile - _cbActual - _cchCRsEaten;
			const unsigned ibVirtual = ibStartBuf + _ibCurrent;

			unsigned ibNewVirtual;
			if (nType == SEEK_CUR)
			{
				// Need signed arithmetic, since relative offset is signed
				ibNewVirtual = (unsigned)((int)ibVirtual + ibOffset);
			}
			else // SEEK_SET
			{
				ibNewVirtual = (unsigned)ibOffset;
			}

			if (ibNewVirtual >= ibStartBuf && ibNewVirtual < ibInFile)
			{
				_ibCurrent = ibNewVirtual - ibStartBuf;
				return TRUE;
			}
		}

		// Untested hypothesis: it is worthwhile to preserve sequential-read behavior
		// when presenting reads and seeks to the filesystem.
		// Adjust the relative seek offset to compensate for read-buffer position,
		// i.e. the characters buffered but not yet "read,"
		// since true file offset is ahead of virtual offset.
		// Could instead convert all relative seeks to absolute seeks,
		// but then the lack of true CRLF adjustment might bite harder.

		if (nType == SEEK_CUR)
		{
			const unsigned dbRelative = _cbActual - _ibCurrent;
			ibOffset -= ((int)dbRelative + (int)_cchCRsEaten);
		}

		// Having adjusted for the seek-ptr offset as necessary, 
		// discard the now useless readbuffer

		ResetReadBuffer();
	}

	if (_fUseUnicode)
	{
		int nGuard = ibOffset; // test for overflow
		ibOffset *= sizeof(WCHAR);

		if (  ((ibOffset > 0) && (ibOffset < nGuard))
			||((ibOffset < 0) && (ibOffset > nGuard)) )
			return FALSE;

		if (dwMethod != FILE_END)
		{
			nGuard = ibOffset;
			ibOffset += sizeof(WCHAR); // for BOM
			if (ibOffset < nGuard) // test again for overflow
				return FALSE;
		}
	}

	if (0xffffffff == ::SetFilePointer(_h, ibOffset, NULL, dwMethod))
		return FALSE;
	return TRUE;
}


void CXFileStream::Close()
{
	ResetReadBuffer();
	ReleaseBuffer();

	if (NULL == _h)
		return;

	::CloseHandle(_h);
	_h = NULL;
}


/* MODE flags for an opened file:
 *  "r": read       "w": write      "a": append
 *  "r+": read/write                "w+": open empty for read/write
 *  "a+": read/append
 * Append "t" or "b" for text and binary mode.
 *  "tu" forces the text-mode to Unicode; "ts" to Ansi.
 */

BOOL CXFileStream::Open(const TCHAR* pszPathName, const char* pszMode)
{
	if (NULL != _h)
		Close();

	BOOL fTruncate = FALSE, fCreateIfNecessary = FALSE;

	// Crack the 'mode' arg
	_fReading = _fWriting = _fAppending = FALSE;
	switch (*pszMode++)
	{
	case 'r':
		_fReading = TRUE;
		fCreateIfNecessary = FALSE;
		break;
	case 'w':
		_fWriting = TRUE;
		fTruncate = TRUE;
		fCreateIfNecessary = TRUE;
		break;
	case 'a':
		_fWriting = TRUE;
		_fAppending = TRUE;
		fTruncate = FALSE;
		fCreateIfNecessary = TRUE;
		break;
	default:
		return FALSE;
	}

	BOOL fForceUnicode = FALSE;
	BOOL fForceANSI = FALSE;
	BOOL fForceCooked = FALSE;
	BOOL fForceRaw = FALSE;

	while (*pszMode)
	{
		switch (*pszMode++)
		{
		case '+':
			if (_fReading) // r
				_fWriting = TRUE;
			else if (_fWriting) // w, a
				_fReading = TRUE;
			break;
		case 't':
			if (fForceRaw)
				return FALSE;
			fForceCooked = TRUE;
			break;
		case 'u':
			if (fForceANSI || fForceRaw)
				return FALSE;
			fForceCooked = TRUE;
			fForceUnicode = TRUE;
			break;
		case 's':
			if (fForceUnicode || fForceRaw)
				return FALSE;
			fForceCooked = TRUE;
			fForceANSI = TRUE;
			break;
		case 'b':
			if (fForceCooked)
				return FALSE;
			fForceRaw = TRUE;
			break;
		default:
			return FALSE;
		}
	}

	// Default read/write modes.
	// Lets i/o to native textfiles (e.g. *.scm) succeed.
	// Pocket Office uses ANSI textfiles! Go figure. Win95 braindamage?

	_fCooked = fForceCooked || !fForceRaw; // if neither, default to cooked
	_fUseUnicode = !(fForceANSI || !fForceUnicode); // if neither, default to ansi

	DWORD dwAccess = 0L;
	if (_fReading)
		dwAccess |= GENERIC_READ;
	if (_fWriting)
		dwAccess |= GENERIC_WRITE;
	DWORD dwShare = FILE_SHARE_READ;
	if (!_fWriting)
		dwShare |= FILE_SHARE_WRITE;
	DWORD dwCreate = 0L;
	if (!fCreateIfNecessary)
		dwCreate = OPEN_EXISTING;
	else if (fTruncate)
		dwCreate = CREATE_ALWAYS;
	else
		dwCreate = OPEN_ALWAYS;

	HANDLE h = ::CreateFile(pszPathName, dwAccess, dwShare, NULL, dwCreate, FILE_ATTRIBUTE_NORMAL, NULL);
	if (INVALID_HANDLE_VALUE == h)
		return FALSE;
	_h = h;

	BOOL fFoundUnicodeBOM = FALSE;

	if (_fCooked && _fReading)
	{
		// For read-only files, the presence or absence of the Unicode start sequence
		// dictates Unicode-mode, rather than the mode flags.

		WCHAR wch;
		DWORD cbReadActual = 0L;

		if (::ReadFile(_h, &wch, sizeof(wch), &cbReadActual, NULL) 
			&& (sizeof(wch) == cbReadActual)
			&& (wch == 0xFEFF))
		{
			// Indeed, file is Unicode.
			_fUseUnicode = TRUE;
			fFoundUnicodeBOM = TRUE;
		}
		else
		{
			::SetFilePointer(_h, 0, NULL, FILE_BEGIN);
			if (!_fWriting)
				_fUseUnicode = FALSE;
		}
	}
	if (_fCooked && _fWriting && _fUseUnicode && !fFoundUnicodeBOM)
	{
		// New cooked-output Unicode files need the Unicode BOM.

		WCHAR wch = 0xFEFF;
		DWORD cbWritten = 0L;

		if (::WriteFile(_h, &wch, sizeof(wch), &cbWritten, NULL) 
			&& (sizeof(wch) == cbWritten))
		{
			// File successfully stamped as Unicode.
			fFoundUnicodeBOM = TRUE;
		}
		else
		{
			::CloseHandle(_h);
			_h = NULL;
			return FALSE;
		}
	}

	return TRUE;
}


FILE *fopen( const char *filename, const char *mode )
{
	for (int i = 0; i < MAXOPENFILES; ++i)
	{
		if (!RgStdioFiles[i].fActive)
		{
			TCHAR szPath[260+1];
			int cch = ::MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, filename, -1,
				szPath, sizeof(szPath)/sizeof(TCHAR));
			if (0 == cch)
				return NULL;

			CXFileStream* pstream = new CXFileStream;
			if (!pstream)
				return NULL;
			if (!pstream->Open(szPath, mode))
			{
				delete pstream;
				return NULL;
			}
			RgStdioFiles[i].pstream = pstream;
			RgStdioFiles[i].fActive = TRUE;
			return &RgStdioFiles[i];
		}
	}
	return NULL;
}


int fflush( FILE *stream )
{
	if (IsInvalidFile(stream))
		return -1;
	stream->pstream->Flush();
	return 0;
}


int fclose( FILE *stream )
{
	if (IsInvalidFile(stream))
		return -1;
	if (stdin == stream || stdout == stream)
		return -1;
	((CXFileStream*)(stream->pstream))->Close();
	delete stream->pstream;
	stream->pstream = NULL;
	stream->fActive = FALSE;
	return 0;
}


size_t fread( void *buffer, size_t size, size_t count, FILE *stream )
{
	if (IsInvalidFile(stream))
		return -1;
	unsigned cbActual;
	stream->pstream->Read((BYTE*)buffer, size*count, &cbActual);
	return (size_t)cbActual;
}


size_t fwrite( const void *buffer, size_t size, size_t count, FILE *stream )
{
	if (IsInvalidFile(stream))
		return -1;
	unsigned cbActual;
	stream->pstream->Write((BYTE*)buffer, size*count, &cbActual);
	return (size_t)cbActual;
}


int getc( FILE *stream )
{
	if (IsInvalidFile(stream))
		return -1;
	unsigned char ch;
	unsigned cbActual = 0;
	stream->pstream->Read(&ch, 1, &cbActual);
	if (cbActual != 1)
		return EOF;
	return (int)ch;
}


int putc( int c, FILE *stream )
{
	if (IsInvalidFile(stream))
		return -1;
	unsigned char ch = (unsigned char)(unsigned int)c;
	unsigned cbActual = 0;
	stream->pstream->Write(&ch, 1, &cbActual);
	if (cbActual != 1)
		return EOF;
	return (int)ch;
}


int ungetc( int c, FILE *stream )
{
	if (IsInvalidFile(stream))
		return -1;
	if (EOF != c)
		stream->pstream->Pushback((unsigned char)(unsigned int)c);
	return c;
}


long ftell( FILE *stream )
{
	if (IsInvalidFile(stream))
		return -1;
	return (long)(stream->pstream->Tell());
}


int fseek( FILE *stream, long offset, int origin )
{
	if (IsInvalidFile(stream))
		return -1;
	if (!stream->pstream->Seek(offset, origin))
		return -1; // error
	return 0;
}


char *fgets( char *string, int n, FILE *stream )
{
	if (IsInvalidFile(stream))
		return NULL;
	char* pch = string;
	int cchRemaining = n-1;
	CXStdioStream *const pstream = stream->pstream;
	while (cchRemaining-- > 0)
	{
		unsigned char ch;
		unsigned cbActual = 0;
		pstream->Read(&ch, 1, &cbActual);
		if (cbActual != 1)
		{
			*pch = 0;
			return NULL;
		}

		*pch++ = ch;
		if ('\n' == ch)
			break;
	}
	*pch = 0;

	return string;
}


int fputs( const char *string, FILE *stream )
{
	if (IsInvalidFile(stream))
		return -1;
	const unsigned cbGiven = strlen(string);
	unsigned cbActual = 0;
	stream->pstream->Write((const BYTE*)string, cbGiven, &cbActual);
	return (cbGiven == cbActual ? 1 : EOF);
}


static int _output(char* buffer, const char* format, va_list argptr)
{
	int cchFormat = strlen(format);
	TCHAR szFormat[128];
	const int cchFormatMax = sizeof(szFormat)/sizeof(TCHAR)-1;

	if (cchFormat > cchFormatMax)
		return 0;
	int i = mbstowcs(szFormat, format, cchFormatMax);
	if (((size_t)-1) == ((size_t)i))
		return 0;
	if (cchFormatMax == i)
		szFormat[cchFormatMax] = 0;
	cchFormat = i;

	// Patch format string to accommodate CE's Unicode sensibilities
	// %s -> %S
	// %c -> %C
	BOOL fFmt; // set when within a format-specifier
	for (i = 0, fFmt = FALSE; i < cchFormatMax && szFormat[i] != 0; ++i)
	{
		switch (szFormat[i])
		{
		case L'%':
			fFmt = !fFmt;
			break;
		case L'0':
		case L'1':
		case L'2':
		case L'3':
		case L'4':
		case L'5':
		case L'6':
		case L'7':
		case L'8':
		case L'9':
		case L'.':
			// Precision/width specifier: part of format
			break;
		case L'c':
			if (fFmt)
			{
				szFormat[i] = L'C';
				fFmt = FALSE;
			}
			break;
		case L's':
			if (fFmt)
			{
				szFormat[i] = L'S';
				fFmt = FALSE;
			}
			break;
		default:
			if (fFmt)
				fFmt = FALSE;
			break;
		}
	}

	// Put a fence on either side of output buffer to detect overflow.
	//$ BUGBUG - no bounds checking on this small buffer.
	// I wish that the CE RTL had _vnswprintf

	volatile int nFence0 = 1;
	TCHAR szOut[256];
	volatile int nFence1 = 1;

	const int cchOutMax = sizeof(szOut)/sizeof(TCHAR)-1;
	// This version from the crt (as opposed to wvsprintf) handles %g
	i = vswprintf(szOut, szFormat, argptr);
	assert(nFence0 == 1 && nFence1 == 1);
	if (nFence0 != 1 || nFence1 != 1)
		return 0;
	if (0 == i)
		return 0;
	if (cchOutMax == i)
		szOut[cchOutMax] = 0;

	// Now convert to mbcs.
	//$ REVIEW - if xstdio had wide-char entry points,
	// the fprintf version could zap this here into WriteW,
	// saving a wcs-to-mbcs-to-wcs round trip

	i = wcstombs(buffer, szOut, cchOutMax);
	if (((size_t)-1) == ((size_t)i))
		return 0;
	if (cchOutMax == i)
		buffer[cchOutMax] = 0;
	return i;
}


int __cdecl sprintf( char *buffer, const char *format, ... )
{
	va_list arglist;
	va_start(arglist, format);
	int i = _output(buffer, format, arglist);
	va_end(arglist);
	return i;
}


int __cdecl fprintf( FILE *stream, const char *format, ... )
{
	if (IsInvalidFile(stream))
		return 0;

	char szTmp[256];
	va_list arglist;
	va_start(arglist, format);
	int i = _output(szTmp, format, arglist);
	va_end(arglist);
	//$ REVIEW - would be good to short-circuit the wcs-mbcs-wcs loop
	fputs(szTmp, stream);
	return i;
}


int __cdecl printf( const char *format, ... )
{
	if (IsInvalidFile(stdout))
		return 0;

	char szTmp[256];
	va_list arglist;
	va_start(arglist, format);
	int i = _output(szTmp, format, arglist);
	va_end(arglist);
	//$ REVIEW - would be good to short-circuit the wcs-mbcs-wcs loop
	fputs(szTmp, stdout);
	return i;
}


int StdioCharReady( FILE *stream )
{
	if (IsInvalidFile(stream))
		return 0;
	return (stream->pstream->CharReady());
}



#endif
